<?php

namespace Drupal\bibcite_import_orcid;

use GuzzleHttp\Client;
use GuzzleHttp\Exception\RequestException;
use Drupal\bibcite_entity\Entity\Reference;
use RenanBr\BibTexParser\Listener;
use RenanBr\BibTexParser\Parser;
use RenanBr\BibTexParser\Processor;
use RenanBr\BibTexParser\Exception\ProcessorException;

/**
 * Import ORCID data.
 */
class Fetch {

  /**
   * Get the all user's works from ORCID Public API.
   * @return array
   */
  public function getAllWorksFromOrcid($orcid) {
    $url = "https://pub.orcid.org/v3.0/{$orcid}/works";
    $client = new Client();
    try {
      $response = $client->get($url, [
        'headers' => [
          'Accept' => 'application/json',
        ]
      ]);
      $works = json_decode($response->getBody(), true);

      return $works['group'];
    } catch (RequestException $e) {
      \Drupal::messenger()->addStatus(t('Processor Exception: ') . $e->getMessage());
      return NULL;
    }
  }

  /**
   * Get a publications details from the ORCID Public API.
   * @return array
   */
  public function getOrcidWorkData($work_put_code, $orcid) {
    $url = "https://pub.orcid.org/v3.0/{$orcid}/works/{$work_put_code}";
    $client = new Client();
    try {
      $response = $client->get($url, [
        'headers' => [
          'Accept' => 'application/json',
        ]
      ]);
      $work_data = json_decode($response->getBody(), true);
      return $work_data["bulk"][0]["work"];
    } catch (RequestException $e) {
      \Drupal::messenger()->addStatus(t('Processor Exception: ') . $e->getMessage());
      return NULL;
    }
  }

  /**
   * Map reference types.
   * @return string
   */
  public function bibcite_entity_types($entity) {
    $pubs = [
      "annotation" => "miscellaneous",
      "artistic-performance" => "miscellaneous",
      "book-chapter" => "book_chapter",
      "book-review" => "miscellaneous",
      "book" => "book",
      "conference-abstract" => "conference_proceedings",
      "conference-paper" => "conference_paper",
      "conference-poster" => "miscellaneous",
      "data-set" => "database",
      "dictionary-entry" => "miscellaneous",
      "disclosure" => "miscellaneous",
      "dissertation-thesis" => "thesis",
      "edited-book" => "book",
      "encyclopedia-entry" => "miscellaneous",
      "invention" => "miscellaneous",
      "journal-article" => "journal_article",
      "journal-issue" => "journal",
      "lecture-speech" => "presentation",
      "license" => "miscellaneous",
      "magazine-article" => "magazine_article",
      "manual" => "miscellaneous",
      "newsletter-article" => "miscellaneous",
      "newspaper-article" => "newspaper_article",
      "online-resource" => "website",
      "other" => "miscellaneous",
      "patent" => "patent",
      "physical-object" => "miscellaneous",
      "presentation" => "presentation",
      "preprint" => "preprint",
      "registered-copyright" => "miscellaneous",
      "report" => "report",
      "research-technique" => "miscellaneous",
      "research-tool" => "miscellaneous",
      "software" => "software",
      "spin-off-company" => "miscellaneous",
      "standards-and-policy" => "miscellaneous",
      "supervised-student-publication" => "miscellaneous",
      "technical-standard" => "miscellaneous",
      "test" => "miscellaneous",
      "trademark" => "miscellaneous",
      "translation" => "miscellaneous",
      "website" => "website",
      "working-paper" => "miscellaneous",
      "undefined" => "miscellaneous",
    ];
    if (array_key_exists($entity, $pubs)) {
      return $pubs[$entity];
    } else {
      return "miscellaneous";
    }
  }

  /**
   * Get the user data from ORCID Public API.
   * @return object
   */
  public function getOrcidPersonData($orcid) {
    $url = "https://pub.orcid.org/v3.0/{$orcid}";
    $client = new Client();
    try {
      $response = $client->get($url, [
        'headers' => [
          'Accept' => 'application/json',
        ]
      ]);
      return json_decode($response->getBody(), true);
    } catch (RequestException $e) {
      \Drupal::messenger()->addStatus(t('Processor Exception: ') . $e->getMessage());
      return NULL;
    }
  }

  /**
   * Get the authors from work, citation or profile.
   * @return array
   */
  public function get_orcid_authors($work_data, $orcid) {
    $bibtex = $work_data["citation"]["citation-value"];
    $bibtex_info = !empty($bibtex) ? $this->parse_bibtex($bibtex) : "";
    $contribs = $work_data['contributors']['contributor'];
    if ($contribs[0]["credit-name"]["value"]) {
      $contributors = array_map(function ($contrib) {
        return $contrib['credit-name']['value'];
      }, $contribs);
    } else if ($bibtex_info) {
      $contribs = $bibtex_info['bibcite_authors'];
      foreach ($contribs as $contributor) {
        $contributors[] = $contributor["first"] . ' ' . $contributor["last"];
      }
    } else {
      $person_data = $this->getOrcidPersonData($orcid);
      $contributors[] = $person_data["person"]["name"]["given-names"]["value"] . ' ' . $person_data["person"]["name"]["family-name"]["value"];
    }
    return $contributors;
  }

  /**
   * Get the authors from work, citation or profile.
   * @return array
   */
  public function get_reference_authors($query, $uid, $contributors) {
    $bibcite_reference = Reference::load(reset($query));
    $pub_refs = $bibcite_reference->author->referencedEntities();
    if (!empty($pub_refs)) {
      // Get the contributors name from the existing publication
      $pub_contributors = [];
      foreach ($pub_refs as $referenced_contributor) {
        $key = $referenced_contributor->id();
        $pub_contributors[$key] = $referenced_contributor->getName();
      };
      // Look for linked authors
      $linked_authors = [];
      foreach ($pub_contributors as $key => $contrib_id) {
        $query_users = \Drupal::entityQuery('user')
          ->condition('field_author', $key)
          ->execute();
        if ($query_users) {
          $linked_authors[$key] = reset($query_users);
        }
      }
    } else {
      foreach ($contributors as $contributor) {
        $options[] = $contributor;
      }
    }

    // Check if the current user is already an author
    foreach ($pub_contributors as $key => $contributor) {
      if (array_key_exists($key, $linked_authors)) {
        $attributes = [];
        // Is the onwer?
        if (in_array($uid, $linked_authors)) {
          $options[$key] = $contributor;
          $attributes[$key] = array('disabled' => 'disabled', 'readonly' => true);
        } else {
          // other linked users
          $options[$key] = $contributor;
          $attributes[$key] = array('disabled' => 'disabled');
        }
      } else {
        // Unlinked users
        $options[$key] = $contributor;
      }
      // Are there authors missing
      if ($pub_contributors && $contributors && count($pub_contributors) < count($contributors)) {
        $missing_contribs = array_diff($contributors, $pub_contributors);
        // Add an option for each contributor
        foreach ($missing_contribs as $key => $contributor) {
          $options[$key] = $contributor;
        }
      }
    }
    return [$options, $attributes];
  }
  /**
   * Process bibtex citation if it exists.
   * @return array
   */
  public function parse_bibtex($work) {
    if (validate_bibtex_syntax($work)) {
      return;
    }
    try {
      $parser = new Parser();
      $listener = new Listener();
      $listener->addProcessor(new Processor\TagNameCaseProcessor(CASE_LOWER));
      $listener->addProcessor(new Processor\NamesProcessor());
      $parser->addListener($listener);
      $parser->parseString($work);
      $reference_data = [];
      $entries = $listener->export();

      foreach ($entries[0] as $prop => $value) {
        switch ($prop) {
          case 'pages':
          case 'Pages':
          case 'PAGES':
            $reference_data['bibcite_pages'] = $entries[0][$prop];
            break;
          case 'number':
          case 'Number':
          case 'NUMBER':
            $reference_data['bibcite_issue'] = $entries[0][$prop];
            break;
          case 'volume':
          case 'Volume':
          case 'VOLUME':
            $reference_data['bibcite_volume'] = $entries[0][$prop];
            break;
          case 'author':
          case 'Author':
          case 'AUTHOR':
            $reference_data['bibcite_authors'] = $entries[0][$prop];
            break;
          case 'booktitle':
          case 'Booktitle':
          case 'BookTitle':
          case 'BOOKTITLE':
            $reference_data['bibcite_booktitle'] = $entries[0][$prop];
            break;
          case 'series':
          case 'Series':
          case 'SERIES':
            $reference_data['bibcite_section'] = $entries[0][$prop];
            break;
          case 'publisher':
          case 'Publisher':
          case 'PUBLISHER':
            $reference_data['bibcite_publisher'] = $entries[0][$prop];
            break;
        }
      }
      return $reference_data;
    } catch (ProcessorException $e) {
      \Drupal::messenger()->addStatus(t('Bibtex Exception: ') . $e->getMessage() . ' ' . $work);
    }
  }
}

/**
 * Checks if bibtex is valid.
 * @return bool
 */
function validate_bibtex_syntax($input) {
  $bibtex_regex = '/@([a-zA-Z]+){([^,]+),\s*((?:[^{}]|{(?:[^{}]|{[^{}]*})*})*)}/';
  if (!preg_match($bibtex_regex, $input)) {
    return false;
  }
  $stack = array();
  for ($i = 0; $i < strlen($input); $i++) {
    $char = $input[$i];
    if ($char == '{') {
      array_push($stack, '{');
    } else if ($char == '}') {
      if (count($stack) == 0) {
        return false;
      } else {
        array_pop($stack);
      }
    }
  }
  if (count($stack) > 0) {
    return false;
  }
  return true;
}
